#
#  Copyright (C) 2000-2026 Greg Landrum and other RDKit contributors
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
""" A module for molecules and stuff

 see Chem/index.html in the doc tree for documentation

"""
from rdkit import DataStructs, RDConfig, rdBase
from rdkit.Chem import rdchem
from rdkit.Geometry import rdGeometry

_HasSubstructMatchStr = rdchem._HasSubstructMatchStr
from rdkit.Chem.inchi import *
from rdkit.Chem.rdchem import *
from rdkit.Chem.rdCIPLabeler import *
from rdkit.Chem.rdmolfiles import *
from rdkit.Chem.rdmolops import *

try:
  # This is an optional component of the build
  from rdkit.Chem.rdMolInterchange import *
except ImportError:
  pass

# Coordgen needs to know where its template file is.
# The default install puts it in RDDataDir
try:
  from rdkit.Chem import rdCoordGen
except ImportError:
  pass
else:
  templDir = RDConfig.RDDataDir
  if templDir[-1] != '/':
    templDir += '/'
  rdCoordGen.SetDefaultTemplateFileDir(templDir)


# Github Issue #6208: boost::python iterators are slower than they should
# (cause is that boost::python throws exceptions as actual C++ exceptions)
class _GetRDKitObjIterator:

  def _sizeCalc(self):
    raise NotImplementedError()

  def _getRDKitItem(self, i):
    raise NotImplementedError()

  def __init__(self, mol):
    self._mol = mol
    self._pos = 0
    self._size = self._sizeCalc()

  def __len__(self):
    if self._sizeCalc() != self._size:
      raise RuntimeError('size changed during iteration')
    return self._size

  def __getitem__(self, i):
    if i < 0 or i >= len(self):
      raise IndexError('index out of range')
    return self._getRDKitItem(i)

  def __next__(self):
    if self._pos >= len(self):
      raise StopIteration

    ret = self[self._pos]
    self._pos += 1
    return ret

  def __iter__(self):
    for i in range(0, len(self)):
      self._pos = i
      yield self[i]
    self._pos = self._size


class _GetAtomsIterator(_GetRDKitObjIterator):

  def _sizeCalc(self):
    return self._mol.GetNumAtoms()

  def _getRDKitItem(self, i):
    return self._mol.GetAtomWithIdx(i)


class _GetBondsIterator(_GetRDKitObjIterator):

  def _sizeCalc(self):
    return self._mol.GetNumBonds()

  def _getRDKitItem(self, i):
    return self._mol.GetBondWithIdx(i)


rdchem.Mol.GetAtoms = lambda self: _GetAtomsIterator(self)
rdchem.Mol.GetAtoms.__doc__ = """returns an iterator over the atoms in the molecule"""
rdchem.Mol.GetBonds = lambda self: _GetBondsIterator(self)
rdchem.Mol.GetBonds.__doc__ = """returns an iterator over the bonds in the molecule"""


def QuickSmartsMatch(smi, sma, unique=True, display=False):
  ''' A convenience function for quickly matching a SMARTS against a SMILES

  Arguments:
    - smi: the SMILES to match
    - sma: the SMARTS to match
    - unique: (optional) determines whether or not only unique matches are returned
    - display: (optional) IGNORED

  Returns:
    a list of list of the indices of the atoms in the molecule that match the SMARTS  
  
  '''
  m = MolFromSmiles(smi)
  p = MolFromSmarts(sma)
  res = m.GetSubstructMatches(p, unique)
  return res


def CanonSmiles(smi, useChiral=1):
  ''' A convenience function for canonicalizing SMILES

  Arguments:
    - smi: the SMILES to canonicalize
    - useChiral: (optional) determines whether or not chiral information is included in the canonicalization and SMILES

  Returns:
    the canonical SMILES

  '''
  m = MolFromSmiles(smi)
  return MolToSmiles(m, useChiral)


def SupplierFromFilename(fileN, delim='', **kwargs):
  ''' A convenience function for creating a molecule supplier from a filename 
  
  Arguments:
    - fileN: the name of the file to read from
    - delim: (optional) the delimiter to use for reading the file (only for csv and txt files)
    - kwargs: additional keyword arguments to be passed to the supplier constructor

  Returns:
    a molecule supplier

  '''
  ext = fileN.split('.')[-1].lower()
  if ext == 'sdf':
    suppl = SDMolSupplier(fileN, **kwargs)
  elif ext == 'csv':
    if not delim:
      delim = ','
    suppl = SmilesMolSupplier(fileN, delimiter=delim, **kwargs)
  elif ext == 'txt':
    if not delim:
      delim = '\t'
    suppl = SmilesMolSupplier(fileN, delimiter=delim, **kwargs)
  elif ext == 'tdt':
    suppl = TDTMolSupplier(fileN, delimiter=delim, **kwargs)
  else:
    raise ValueError("unrecognized extension: %s" % ext)

  return suppl


def FindMolChiralCenters(mol, force=True, includeUnassigned=False, includeCIP=True,
                         useLegacyImplementation=None):
  """ returns information about the chiral centers in a molecule

  Arguments:
    - mol: the molecule to work with
    - force: (optional) if True, stereochemistry will be assigned even if it has been already
    - includeUnassigned: (optional) if True, unassigned stereo centers will be included in the output
    - includeCIP: (optional) if True, the CIP code for each chiral center will be included in the output
    - useLegacyImplementation: (optional) if True, the legacy stereochemistry perception code will be used

  Returns:
    a list of tuples of the form (atomId, CIPCode)

    >>> from rdkit import Chem
    >>> mol = Chem.MolFromSmiles('[C@H](Cl)(F)Br')
    >>> Chem.FindMolChiralCenters(mol)
    [(0, 'R')]
    >>> mol = Chem.MolFromSmiles('[C@@H](Cl)(F)Br')
    >>> Chem.FindMolChiralCenters(mol)
    [(0, 'S')]

    >>> Chem.FindMolChiralCenters(Chem.MolFromSmiles('CCC'))
    []

    By default unassigned stereo centers are not reported:

    >>> mol = Chem.MolFromSmiles('C[C@H](F)C(F)(Cl)Br')
    >>> Chem.FindMolChiralCenters(mol,force=True)
    [(1, 'S')]

    but this can be changed:

    >>> Chem.FindMolChiralCenters(mol,force=True,includeUnassigned=True)
    [(1, 'S'), (3, '?')]

    The handling of unassigned stereocenters for dependent stereochemistry is not correct 
    using the legacy implementation:

    >>> Chem.FindMolChiralCenters(Chem.MolFromSmiles('C1CC(C)C(C)C(C)C1'),includeUnassigned=True, useLegacyImplementation=True)
    [(2, '?'), (6, '?')]
    >>> Chem.FindMolChiralCenters(Chem.MolFromSmiles('C1C[C@H](C)C(C)[C@H](C)C1'),includeUnassigned=True, useLegacyImplementation=True)
    [(2, 'S'), (4, '?'), (6, 'R')]

    But works with the new implementation:

    >>> Chem.FindMolChiralCenters(Chem.MolFromSmiles('C1CC(C)C(C)C(C)C1'),includeUnassigned=True, useLegacyImplementation=False)
    [(2, '?'), (4, '?'), (6, '?')]

    Note that the new implementation also gets the correct descriptors for para-stereochemistry:

    >>> Chem.FindMolChiralCenters(Chem.MolFromSmiles('C1C[C@H](C)[C@H](C)[C@H](C)C1'),useLegacyImplementation=False)
    [(2, 'S'), (4, 's'), (6, 'R')]

    With the new implementation, if you don't care about the CIP labels of stereocenters, you can save
    some time by disabling those:

    >>> Chem.FindMolChiralCenters(Chem.MolFromSmiles('C1C[C@H](C)[C@H](C)[C@H](C)C1'), includeCIP=False, useLegacyImplementation=False)
    [(2, 'Tet_CCW'), (4, 'Tet_CCW'), (6, 'Tet_CCW')]

  """
  origUseLegacyVal = GetUseLegacyStereoPerception()
  if useLegacyImplementation is None:
    useLegacyImplementation = origUseLegacyVal
  SetUseLegacyStereoPerception(useLegacyImplementation)
  try:
    if useLegacyImplementation:
      for atom in mol.GetAtoms():
        atom.ClearProp('_ChiralityPossible')
      AssignStereochemistry(mol, force=force, flagPossibleStereoCenters=includeUnassigned,
                            cleanIt=True)
      centers = []
      for atom in mol.GetAtoms():
        if atom.HasProp('_CIPCode'):
          centers.append((atom.GetIdx(), atom.GetProp('_CIPCode')))
        elif includeUnassigned and atom.HasProp('_ChiralityPossible'):
          centers.append((atom.GetIdx(), '?'))
    else:
      centers = []
      itms = FindPotentialStereo(mol)
      if includeCIP:
        atomsToLabel = []
        bondsToLabel = []
        for si in itms:
          if si.type == StereoType.Atom_Tetrahedral:
            atomsToLabel.append(si.centeredOn)
          elif si.type == StereoType.Bond_Double:
            bondsToLabel.append(si.centeredOn)
        AssignCIPLabels(mol, atomsToLabel=atomsToLabel, bondsToLabel=bondsToLabel)
      for si in itms:
        if si.type == StereoType.Atom_Tetrahedral and (includeUnassigned or si.specified
                                                       == StereoSpecified.Specified):
          idx = si.centeredOn
          atm = mol.GetAtomWithIdx(idx)
          if includeCIP and atm.HasProp("_CIPCode"):
            code = atm.GetProp("_CIPCode")
          else:
            if si.specified:
              code = str(si.descriptor)
            else:
              code = '?'
              atm.SetIntProp('_ChiralityPossible', 1)
          centers.append((idx, code))
  finally:
    SetUseLegacyStereoPerception(origUseLegacyVal)
  return centers


def MultiConfMolFromSDF(sdf, sanitize=True, removeHs=True, strictParsing=True):
  """ A convenience function for creating a multi conformer molecule from an SD file.

  This assumes that all structures in the same file are actually the same molecule.
  
  Arguments:
    - sdf: the name of the file to read from
    - sanitize: (optional) Sanitize the molecule after constructing it
    - removeHs: (optional) Remove Hs after constructing the molecule
    - strictParsing: (optional) If set to false, the parser is more lax about the correctness of the contents

  Returns:
    A molecule with all conformers found in the SD file.

  """
  returnMol = None
  for mol in SDMolSupplier(sdf, sanitize=sanitize, removeHs=removeHs, strictParsing=strictParsing):
    if returnMol is None:
      returnMol = Mol(mol)
      continue
    returnMol.AddConformer(mol.GetConformer(), assignId=True)
  return returnMol



#!/usr/bin/env python
# Simple script to dump a TTF file into an array of unsigned chars that
# can be read into FreeType.

import argparse

parser = argparse.ArgumentParser(description='Dump TTF file to char array.')
parser.add_argument('--ttf-file', required=True, help='Name of TTF file.')
parser.add_argument('--output-file', required=True, help='Name of output file.')
parser.add_argument('--variable-name', default='raw_data', help='Name of variable for char array.'
                    '  Default=%(default)s.')
parser.add_argument('--string-name', default='ttf_font_data',
                    help='Name of variable for string array.'
                    '  Default=%(default)s.')
args = parser.parse_args()

with open(args.ttf_file, 'rb') as f:
  hexdata = f.read().hex()
num = 0

with open(args.output_file, 'w') as f:
  f.write('namespace {\n')
  f.write(f'const unsigned char {args.variable_name}[] = {{\n   ')
  for i in range(0, len(hexdata), 2):
    f.write(f' 0x{hexdata[i:i+2]},')
    num += 1
    if num == 12:
      f.write('\n   ')
      num = 0
  f.write('\n};\n')
  f.write('}  // namespace\n')
  f.write(f'const std::string {args.string_name}((const char *){args.variable_name},'
          f' (const char *){args.variable_name} + sizeof({args.variable_name}));')



#!/usr/bin/env python

from json import dumps

from rdkit import Chem, rdBase
from rdkit.Chem import AllChem, Draw, rdDepictor
from rdkit.Chem.Draw import rdMolDraw2D

COLS = [(1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0), (1.0, 0.55, 1.0)]


def get_hit_atoms_and_bonds(mol, smt):
  alist = []
  blist = []
  q = Chem.MolFromSmarts(smt)
  for match in mol.GetSubstructMatches(q):
    alist.extend(match)

  for ha1 in alist:
    for ha2 in alist:
      if ha1 > ha2:
        b = mol.GetBondBetweenAtoms(ha1, ha2)
        if b:
          blist.append(b.GetIdx())

  return alist, blist


def add_colours_to_map(els, cols, col_num):
  for el in els:
    if el not in cols:
      cols[el] = []
    if COLS[col_num] not in cols[el]:
      cols[el].append(COLS[col_num])


def do_a_picture(smi, smarts, filename, label, fmt='svg'):

  with rdDepictor.UsingCoordGen(True):
    mol = Chem.MolFromSmiles(smi)
    mol = Draw.PrepareMolForDrawing(mol)

    acols = {}
    bcols = {}
    h_rads = {}
    h_lw_mult = {}

    for i, smt in enumerate(smarts):
      alist, blist = get_hit_atoms_and_bonds(mol, smt)
      col = i % 4
      add_colours_to_map(alist, acols, col)
      add_colours_to_map(blist, bcols, col)

    if fmt == 'svg':
      d = rdMolDraw2D.MolDraw2DSVG(300, 300)
      mode = 'w'
    elif fmt == 'png':
      d = rdMolDraw2D.MolDraw2DCairo(300, 300)
      mode = 'wb'
    else:
      print('unknown format {}'.format(fmt))
      return

    d.drawOptions().fillHighlights = False
    d.DrawMoleculeWithHighlights(mol, label, acols, bcols, h_rads, h_lw_mult, -1)
    d.FinishDrawing()

    with open(filename, mode) as f:
      f.write(d.GetDrawingText())


smi = 'CO[C@@H](O)C1=C(O[C@H](F)Cl)C(C#N)=C1ONNC[NH3+]'
smarts = ['CONN', 'N#CC~CO', 'C=CON', 'CONNCN']
do_a_picture(smi, smarts, 'atom_highlights_3.png', '', fmt='png')



# -*- coding: utf-8 -*-
"""
Implementation of Ertl's Functional Group Algorithm in RDKit

Ertl, Peter. 

An algorithm to identify functional groups in organic molecules

J Cheminform (2017) 9:36; DOI: 10.1186/s13321-017-0225-z

@author: gonzalo.colmenarejo
"""


from rdkit import Chem
from rdkit.Chem.Draw import IPythonConsole
from chembl_structure_pipeline import standardizer as sdz
from datetime import datetime
from IPython.display import Image
from PIL import Image as pilImage
from PIL import ImageDraw, ImageFont
from rdkit.Chem.Draw import rdMolDraw2D
import efgs
import pandas as pd
import io



ertl_smis = [
    'Cc1nc(NS(=O)(=O)c2ccc(N)cc2)nc(C)c1', 
    'NC(=N)c1ccc(C=Cc2ccc(cc2O)C(=N)N)cc1', 
    'CC(=O)Nc1nnc(s1)S(=O)(=O)N',
    'NS(=O)(=O)c1cc2c(NCNS2(=O)=O)cc1Cl',
    'CNC1=Nc2ccc(Cl)cc2C(=N(=O)C1)c3ccccc3',
    'Cc1onc(c1C(=O)NC2C3SC(C)(C)C(N3C2=O)C(=O)O)c4ccccc4',
    'Clc1ccccc1C2=NCC(=O)Nc3ccc(cc23)N(=O)=O',
    'COc1cc(cc(C(=O)NCC2CCCN2CC=C)c1OC)S(=O)(=O)N',
    'Cc1ccc(Cl)c(Nc2ccccc2C(=O)O)c1Cl',
    'Clc1ccc2Oc3ccccc3N=C(N4CCNCC4)c2c1',
    'FC(F)(F)CN1C(=O)CN=C(c2ccccc2)c3cc(Cl)ccc13',
    'OCC1OC(CC1O)n2cnc3C(O)CNC=Nc32',
    'CCNC1CC(C)S(=O)(=O)c2sc(cc12)S(=O)(=O)N',
    'CC(O)C1C2C(C)C(=C(N2C1=O)C(=O)O)SC3CNC(C3)C(=O)N(C)C',
    'CC1CN(CC(C)N1)c2c(F)c(N)c3c(=O)c(cn(C4CC4)c3c2F)C(=O)O',
    'CC(=CCC1C(=O)N(N(C1=O)c2ccccc2)c3ccccc3)C',
    'Clc1ccc2N=C3NC(=O)CN3Cc2c1Cl',
    'CC(=O)NC1C(NC(=N)N)C=C(OC1C(O)C(O)CO)C(=O)O',
    'CC(O)C(O)C1CNc2nc(N)nc(O)c2N1',
    'NC1CCCCN(C1)c2c(Cl)cc3c(=O)c(cn(C4CC4)c3c2Cl)C(=O)O',
]



def mindex(mol):
    # Utility function to draw molecule with idx
    for atom in mol.GetAtoms():
        try:
            atom.SetAtomMapNum(atom.GetIdx())
        except:
            None
    return mol


def demindex(mol):
    # Utility function to return mindex-ed molecule to normal
    for atom in mol.GetAtoms():
        atom.ClearProp('molAtomMapNumber')
    return mol


def molimg(x):
    # Utility function to draw image of labeled mol from PNG binary string
    try:
        return pilImage.open(io.BytesIO(x))
    except:
        return None
    

def img_grid(images, titles = None, num_columns = 5,font = "arial.ttf", font_size = 18, title_height = 30, cell_width = 300, cell_height = 300, bg_color = (255, 255, 255)):
    """
    Creates a grid of images with optional titles under each image.

    Parameters:
    - images (list): List of PIL Image objects.
    - titles (list, optional): List of titles as strings for each image. If None, no titles will be added.
    - num_columns (int): Number of columns in the grid.
    - font_sizes: Font size of titles (if provided).
    - title_height: Height of titles (if provieded).
    - cell_width (int): Width of each image cell.
    - cell_height (int): Height of each image cell (not including title space).
    - bg_color (tuple): Background color as an RGB tuple.

    Returns:
    - PIL Image object containing the image grid.
    """
    # Set title height if titles are provided; otherwise, set to zero
    title_height = title_height if titles else 0
    cell_height_with_title = cell_height + title_height + 2

    # Calculate grid dimensions
    num_images = len(images)
    num_rows = (num_images + num_columns - 1) // num_columns  # Round up for last row

    # Create a blank background image for the grid
    grid_width = num_columns * cell_width
    grid_height = num_rows * cell_height_with_title
    grid_image = pilImage.new("RGB", (grid_width, grid_height), bg_color)

    # Set up font for titles if needed
    if titles:
        try:
            font = ImageFont.truetype(font, font_size)  # Change font and size as desired
        except IOError:
            font = ImageFont.load_default()

    # Iterate through each cell in the grid
    for i in range(num_rows * num_columns):
        row = i // num_columns
        col = i % num_columns
        x = col * cell_width
        y = row * cell_height_with_title

        if i < num_images:
            # Resize the image to fit in the cell
            img_resized = images[i].resize((cell_width, cell_height))
            grid_image.paste(img_resized, (x, y))

            # Draw title if provided
            if titles and i < len(titles):
                draw = ImageDraw.Draw(grid_image)
                title = titles[i]

                # Calculate title position and center it
                text_bbox = draw.textbbox((0, 0), title, font=font)
                text_width = text_bbox[2] - text_bbox[0]
                title_x = x + (cell_width - text_width) // 2  # Center title horizontally
                title_y = y + cell_height  # Position directly beneath the image

                # Draw title text
                draw.text((title_x, title_y), title, fill="black", font=font)
        else:
            # Leave blank space if no image for this cell
            blank_cell = pilImage.new("RGB", (cell_width, cell_height), bg_color)
            grid_image.paste(blank_cell, (x, y))

    return grid_image



# ## Mols from file
start_time = datetime.now()
d = pd.read_csv("../data/ch33_inchis.csv", sep = ";", low_memory = False)
d = d[pd.notna(d.inchi)].reset_index(drop = True) # make sure there are no missing inchis
d = d[d.inchi.apply(lambda x: "." not in x.split("/")[1])].reset_index(drop = True) # make sure we have no complex molecules
d.shape # (839063, 1)
d = d.sample(frac=1, random_state=4210).reset_index(drop = True) # scramble the df
n = d.shape[0] # Here one can put an alternative much smaller n 
d = d[d.index.isin(range(n))].reset_index(drop = True) # keep only n
d["mol"] = d.inchi.apply(lambda x: sdz.standardize_mol(Chem.MolFromInchi(x)))
d = d[pd.notna(d.mol)].reset_index(drop = True) # remove entries without molecule object
d.shape[0]  # n
end_time = datetime.now()
print('Duration to analyze {} molecules: {}'.format(n, end_time - start_time))

start_time = datetime.now()
d[["img_text","fgs","fgsmi","fgmol"]] = d.apply(lambda x: pd.Series(efgs.get_dec_fgs(x["mol"])), axis = 1)
end_time = datetime.now()
print('Duration to analyze {} molecules: {}'.format(n, end_time - start_time))


# Generate distribution of FGs
fg_dist = d.explode(["fgsmi","fgmol"]).groupby("fgsmi", as_index = False).apply(lambda x: 
                                                                                pd.DataFrame({"fgsmi": [x["fgsmi"].iloc[0]], 
                                                                                              "n": [x.shape[0]], 
                                                                                              "mol": [x["fgmol"].iloc[0]]})).sort_values("n", ascending = False)

fg_dist.shape  # 5829, 3 (for all ChEMBL comps)
fg_dist.to_csv("../results/fg_dist.csv", sep = ";", index = False)  
# fg_dist = pd.read_csv("../results/fg_dist.csv", sep = ";", index = False) # To recover data
    
mols = fg_dist.mol.iloc[:64].tolist()
legs = fg_dist.fgsmi.iloc[:64].tolist()


## High-resolution grid with top-48 FGs 
images = []
img_width = 250
img_height = 250
for m in mols[:48]:
    Chem.rdDepictor.Compute2DCoords(m)
    Chem.rdDepictor.StraightenDepiction(m)
    d2d = rdMolDraw2D.MolDraw2DCairo(img_width,img_height)
    dopts = d2d.drawOptions()
    dopts.bondLineWidth = 5
    dopts.padding = 0.2
    d2d.DrawMolecule(m)
    d2d.FinishDrawing()
    images.append(molimg(d2d.GetDrawingText()))
   
fgs_grid = img_grid(images[:48], titles = legs[:48], num_columns = 8, font_size = 28, cell_width = 300, cell_height = 300, bg_color=(255, 255, 255))
fgs_grid = fgs_grid.resize((6000, int(6000*fgs_grid.height/fgs_grid.width)))  
fgs_grid.save("../results/fg_grid.png", dpi = (300,300))


# High-resolution grid with 20 first molecules with colored functional groups
images = [molimg(x) for x in d.img_text.iloc[0:20].tolist()]
cell_width = images[0].width
cell_height = images[0].height
molfg_grid = img_grid(images, num_columns = 5, cell_width = cell_width, cell_height = cell_height, bg_color=(255, 255, 255))
molfg_grid
molfg_grid = molfg_grid.resize((6000, int(6000*molfg_grid.height/molfg_grid.width)))
molfg_grid.save("../RESULTS/molfg_grid.png", dpi = (300,300))


# High-resolution grid with 20 Ertl molecules in Figure 1
ertmols = [Chem.MolFromSmiles(s) for s in ertl_smis[:20]]
images = [molimg(efgs.get_dec_fgs(ertmol)[0]) for ertmol in ertmols] 
cell_width = images[0].width
cell_height = images[0].height
molfg_grid = img_grid(images, num_columns = 5, cell_width = cell_width, cell_height = cell_height, bg_color=(255, 255, 255))
molfg_grid
molfg_grid = molfg_grid.resize((6000, int(6000*molfg_grid.height/molfg_grid.width)))
molfg_grid.save("../RESULTS/ertl_molfg_grid.png", dpi = (300,300))



#  $Id$
#
#  Copyright (C) 2003 Rational Discovery LLC
#     All Rights Reserved
#
from rdkit import Chem
from rdkit.VLib.Supply import SupplyNode


class SmilesSupplyNode(SupplyNode):
  """ Smiles supplier

    Sample Usage:
      >>> import os
      >>> from rdkit import RDConfig
      >>> fileN = os.path.join(RDConfig.RDCodeDir,'VLib','NodeLib',\
                               'test_data','pgp_20.txt')
      >>> suppl = SmilesSupplyNode(fileN,delim="\\t",smilesColumn=2,nameColumn=1,titleLine=1)
      >>> ms = [x for x in suppl]
      >>> len(ms)
      20
      >>> ms[0].GetProp("_Name")
      'ALDOSTERONE'
      >>> ms[0].GetProp("ID")
      'RD-PGP-0001'
      >>> ms[1].GetProp("_Name")
      'AMIODARONE'
      >>> ms[3].GetProp("ID")
      'RD-PGP-0004'
      >>> suppl.reset()
      >>> suppl.next().GetProp("_Name")
      'ALDOSTERONE'
      >>> suppl.next().GetProp("_Name")
      'AMIODARONE'
      >>> suppl.reset()

    """

  def __init__(self, fileName, delim="\t", nameColumn=1, smilesColumn=0, titleLine=0, **kwargs):
    SupplyNode.__init__(self, **kwargs)
    self._fileName = fileName
    self._supplier = Chem.SmilesMolSupplier(self._fileName, delimiter=delim,
                                            smilesColumn=smilesColumn, nameColumn=nameColumn,
                                            titleLine=titleLine)

  def reset(self):
    SupplyNode.reset(self)
    self._supplier.reset()

  def next(self):
    """

        """
    r = None
    while not r:
      r = next(self._supplier)
    return r


SmilesSupplyNode.__next__ = SmilesSupplyNode.next


# ------------------------------------
#
#  doctest boilerplate
#
def _runDoctests(verbose=None):  # pragma: nocover
  import doctest
  import sys
  failed, _ = doctest.testmod(optionflags=doctest.ELLIPSIS, verbose=verbose)
  sys.exit(failed)


if __name__ == '__main__':  # pragma: nocover
  _runDoctests()



#
#  Copyright (c) 2010, Novartis Institutes for BioMedical Research Inc.
#  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#     * Neither the name of Novartis Institutes for BioMedical Research Inc.
#       nor the names of its contributors may be used to endorse or promote
#       products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Created by Greg Landrum, October 2006
#
import os
import re
import weakref

from rdkit import Chem, RDConfig


class FGHierarchyNode(object):
  children = None
  name = ""
  label = ""
  pattern = None
  smarts = ""
  rxnSmarts = ""
  parent = None
  removalReaction = None

  def __init__(self, name, patt, smarts="", label="", rxnSmarts="", parent=None):
    self.name = name
    self.pattern = patt
    if parent:
      self.parent = weakref.ref(parent)
    self.label = label
    self.smarts = smarts
    self.children = []
    self.rxnSmarts = rxnSmarts

  def __len__(self):
    res = 1
    for child in self.children:
      res += len(child)
    return res


class FuncGroupFileParseError(ValueError):
  pass


groupDefns = {}
hierarchy = None
lastData = None
lastFilename = None


def BuildFuncGroupHierarchy(fileNm=None, data=None, force=False):
  global groupDefns, hierarchy, lastData, lastFilename
  if (not force and hierarchy and (not data or data == lastData)
      and (not fileNm or fileNm == lastFilename)):
    return hierarchy[:]
  lastData = data
  splitter = re.compile('\t+')

  if not fileNm and not data:
    fileNm = os.path.join(RDConfig.RDDataDir, 'Functional_Group_Hierarchy.txt')

  if fileNm:
    with open(fileNm, 'r') as inF:
      data = inF.readlines()
    lastFilename = fileNm
  elif data:
    data = data.splitlines()
  else:
    raise ValueError("need data or filename")

  groupDefns = {}
  res = []
  for lineNo, line in enumerate(data, 1):
    line = line.strip()
    line = line.split('//')[0]
    if not line:
      continue
    splitL = splitter.split(line)
    if len(splitL) < 3:
      raise FuncGroupFileParseError("Input line %d (%s) is not long enough." % (lineNo, repr(line)))
    label = splitL[0].strip()
    if label in groupDefns:
      raise FuncGroupFileParseError("Duplicate label on line %d." % lineNo)
    labelHierarchy = label.split('.')
    if len(labelHierarchy) > 1:
      for i in range(len(labelHierarchy) - 1):
        tmp = '.'.join(labelHierarchy[:i + 1])
        if tmp not in groupDefns:
          raise FuncGroupFileParseError("Hierarchy member %s (line %d) not found." % (tmp, lineNo))
      parent = groupDefns['.'.join(labelHierarchy[:-1])]
    else:
      parent = None
    smarts = splitL[1]
    patt = Chem.MolFromSmarts(smarts)
    if not patt:
      raise FuncGroupFileParseError('Smarts "%s" (line %d) could not be parsed.' % (smarts, lineNo))

    name = splitL[2].strip()

    rxnSmarts = ''
    if len(splitL) > 3:
      rxnSmarts = splitL[3]

    node = FGHierarchyNode(name, patt, smarts=smarts, label=label, parent=parent,
                           rxnSmarts=rxnSmarts)
    if parent:
      parent.children.append(node)
    else:
      res.append(node)
    groupDefns[label] = node
  hierarchy = res[:]
  return res


def _SetNodeBits(mol, node, res, idx):
  ms = mol.GetSubstructMatches(node.pattern)
  count = 0
  seen = {}
  for m in ms:
    if m[0] not in seen:
      count += 1
      seen[m[0]] = 1
  if count:
    res[idx] = count
    idx += 1
    for child in node.children:
      idx = _SetNodeBits(mol, child, res, idx)
  else:
    idx += len(node)
  return idx


def CreateMolFingerprint(mol, hierarchy):
  totL = 0
  for entry in hierarchy:
    totL += len(entry)
  res = [0] * totL
  idx = 0
  for entry in hierarchy:
    idx = _SetNodeBits(mol, entry, res, idx)
  return res



#  Copyright (C) 2004-2006 Rational Discovery LLC
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
"""
Contains an implementation of Topological-torsion fingerprints, as
described in:

R. Nilakantan, N. Bauman, J. S. Dixon, R. Venkataraghavan;
"Topological Torsion: A New Molecular Descriptor for SAR Applications.
Comparison with Other Descriptors" JCICS 27, 82-85 (1987).

The fingerprints can be accessed through the following functions:
- GetTopologicalTorsionFingerprint
- GetHashedTopologicalTorsionFingerprint
- GetTopologicalTorsionFingerprintAsIntVect (identical to GetTopologicalTorsionFingerprint)
- GetTopologicalTorsionFingerprintAsIds

"""
from rdkit.Chem import rdMolDescriptors
from rdkit.Chem.AtomPairs import Utils
from rdkit.Chem.rdMolDescriptors import (
    GetHashedTopologicalTorsionFingerprint, GetTopologicalTorsionFingerprint)

GetTopologicalTorsionFingerprintAsIntVect = rdMolDescriptors.GetTopologicalTorsionFingerprint


def pyScorePath(mol, path, size, atomCodes=None):
  """ Returns a score for an individual path.

  >>> from rdkit import Chem
  >>> m = Chem.MolFromSmiles('CCCCC')
  >>> c1 = Utils.GetAtomCode(m.GetAtomWithIdx(0), 1)
  >>> c2 = Utils.GetAtomCode(m.GetAtomWithIdx(1), 2)
  >>> c3 = Utils.GetAtomCode(m.GetAtomWithIdx(2), 2)
  >>> c4 = Utils.GetAtomCode(m.GetAtomWithIdx(3), 1)
  >>> t = c1 | (c2 << rdMolDescriptors.AtomPairsParameters.codeSize) | (c3 << (rdMolDescriptors.AtomPairsParameters.codeSize * 2)) | (c4 << (rdMolDescriptors.AtomPairsParameters.codeSize * 3))
  >>> pyScorePath(m, (0, 1, 2, 3), 4) == t
  1

  The scores are path direction independent:

  >>> pyScorePath(m, (3, 2, 1, 0), 4) == t
  1

  >>> m = Chem.MolFromSmiles('C=CC(=O)O')
  >>> c1 = Utils.GetAtomCode(m.GetAtomWithIdx(0), 1)
  >>> c2 = Utils.GetAtomCode(m.GetAtomWithIdx(1), 2)
  >>> c3 = Utils.GetAtomCode(m.GetAtomWithIdx(2), 2)
  >>> c4 = Utils.GetAtomCode(m.GetAtomWithIdx(4), 1)
  >>> t = c1 | (c2 << rdMolDescriptors.AtomPairsParameters.codeSize) | (c3 << (rdMolDescriptors.AtomPairsParameters.codeSize * 2)) | (c4 << (rdMolDescriptors.AtomPairsParameters.codeSize * 3))
  >>> pyScorePath(m, (0, 1, 2, 4), 4) == t
  1

  """
  codes = [None] * size
  for i in range(size):
    if i == 0 or i == size - 1:
      sub = 1
    else:
      sub = 2
    if not atomCodes:
      codes[i] = Utils.GetAtomCode(mol.GetAtomWithIdx(path[i]), sub)
    else:
      codes[i] = atomCodes[path[i]] - sub

  # "canonicalize" the code vector:
  beg = 0
  end = len(codes) - 1
  while beg < end:
    if codes[beg] == codes[end]:
      beg += 1
      end -= 1
    else:
      if codes[beg] > codes[end]:
        codes.reverse()
      break

  accum = 0
  codeSize = rdMolDescriptors.AtomPairsParameters.codeSize
  for i, code in enumerate(codes):
    accum |= code << (codeSize * i)
  return accum


def ExplainPathScore(score, size=4):
  """

  >>> from rdkit import Chem
  >>> m = Chem.MolFromSmiles('C=CC')
  >>> score=pyScorePath(m, (0, 1, 2), 3)
  >>> ExplainPathScore(score, 3)
  (('C', 1, 0), ('C', 2, 1), ('C', 1, 1))

  Again, it's order independent:

  >>> score = pyScorePath(m, (2, 1, 0), 3)
  >>> ExplainPathScore(score, 3)
  (('C', 1, 0), ('C', 2, 1), ('C', 1, 1))

  >>> m = Chem.MolFromSmiles('C=CO')
  >>> score=pyScorePath(m, (0, 1, 2), 3)
  >>> ExplainPathScore(score, 3)
  (('C', 1, 1), ('C', 2, 1), ('O', 1, 0))

  >>> m = Chem.MolFromSmiles('OC=CO')
  >>> score=pyScorePath(m, (0, 1, 2, 3), 4)
  >>> ExplainPathScore(score, 4)
  (('O', 1, 0), ('C', 2, 1), ('C', 2, 1), ('O', 1, 0))

  >>> m = Chem.MolFromSmiles('CC=CO')
  >>> score=pyScorePath(m, (0, 1, 2, 3), 4)
  >>> ExplainPathScore(score, 4)
  (('C', 1, 0), ('C', 2, 1), ('C', 2, 1), ('O', 1, 0))


  >>> m = Chem.MolFromSmiles('C=CC(=O)O')
  >>> score=pyScorePath(m, (0, 1, 2, 3), 4)
  >>> ExplainPathScore(score, 4)
  (('C', 1, 1), ('C', 2, 1), ('C', 3, 1), ('O', 1, 1))
  >>> score=pyScorePath(m, (0, 1, 2, 4), 4)
  >>> ExplainPathScore(score, 4)
  (('C', 1, 1), ('C', 2, 1), ('C', 3, 1), ('O', 1, 0))


  >>> m = Chem.MolFromSmiles('OOOO')
  >>> score=pyScorePath(m, (0, 1, 2), 3)
  >>> ExplainPathScore(score, 3)
  (('O', 1, 0), ('O', 2, 0), ('O', 2, 0))
  >>> score=pyScorePath(m, (0, 1, 2, 3), 4)
  >>> ExplainPathScore(score, 4)
  (('O', 1, 0), ('O', 2, 0), ('O', 2, 0), ('O', 1, 0))

  """
  codeSize: int = rdMolDescriptors.AtomPairsParameters.codeSize
  codeMask = (1 << codeSize) - 1
  res = [None] * size
  for i in range(size):
    if i == 0 or i == size - 1:
      sub = 1
    else:
      sub = 2
    code = score & codeMask
    score = score >> codeSize
    symb, nBranch, nPi = Utils.ExplainAtomCode(code)
    res[i] = (symb, nBranch + sub, nPi)
  return tuple(res)


def GetTopologicalTorsionFingerprintAsIds(mol, targetSize=4):
  nonZeroElements = GetTopologicalTorsionFingerprint(mol, targetSize).GetNonzeroElements()
  frequencies = sorted(nonZeroElements.items())
  res = []
  for k, v in frequencies:
    res.extend([k] * v)
  return res


# ------------------------------------
#
#  doctest boilerplate
#
def _runDoctests(verbose=None):  # pragma: nocover
  import doctest
  import sys
  failed, _ = doctest.testmod(optionflags=doctest.ELLIPSIS, verbose=verbose)
  sys.exit(failed)


if __name__ == '__main__':  # pragma: nocover
  _runDoctests()



#!/usr/bin/env python
#
# File to build a database of fragment scores for the Synthetic
# Accessibility score of Ertl and Schuffenhauer
# Journal of Cheminformatics 1:8 (2009)
# http://www.jcheminf.com/content/1/1/8
#
# It's so you can update the fragment scores to reflect improvements
# in chemistry.  For example, building with a recent SureChEMBL dataset
# substantially reduced all the scores for the original training compounds.

import argparse
import gzip
import math
import pickle

try:
  import seaborn.objects as so
  HAVE_SEABORN = True
except ModuleNotFoundError:
  HAVE_SEABORN = False
import sys

from collections import defaultdict
from pathlib import Path

from rdkit import Chem
from rdkit.Chem import rdFingerprintGenerator


def parse_args(cli_args):
  """
        Use argparse module to parse the command line arguments.

        Returns:
            (Namespace object): Parsed arguments in argparse Namespace
                                object.
        """
  parser = argparse.ArgumentParser(description='Generate fragment scores'
                                   ' data for Synthetic'
                                   ' Accessibility (SA) Score.')
  parser.add_argument('-S', '--score-file', dest='score_file', required=True,
                      help='Name of the output scores file.')
  in_file_args = parser.add_mutually_exclusive_group(required=True)
  in_file_args.add_argument(
    '-I', '--input-file', dest='input_file', help='Name of molecule file containing input'
    ' molecules to be processed.')
  in_file_args.add_argument('--input-freqs-pickle', dest='in_freq_pickle',
                            help='Name of previously generated frequencies'
                            ' pickle file.')
  parser.add_argument(
    '-P', '--plot-file', dest='plot_file', default='dist.svg',
    help='Name of file for plot of distribution of'
    ' frequencies. Must use an extension recognised by'
    ' matplotlib.  Default=%(default)s.')
  parser.add_argument(
    '--output-freqs-pickle', dest='out_freq_pickle',
    help='If this is given, the raw frequencies from the'
    ' molecules will be pickled to this file.')

  args = parser.parse_args(cli_args)
  return args


def create_mol_supplier(infile):
  """

    Args:
        infile (str): must end .smi, .sdf or .sdf.gz

    Returns:
        ForwardSDMolSupplier or None
    """
  inpath = Path(infile)
  sfx = inpath.suffix
  gzipped = False
  if sfx == '.gz':
    suffixes = inpath.suffixes
    gzipped = True
    sfx = suffixes[-2]

  if sfx != '.smi' and sfx != '.sdf' and sfx != '.mol':
    print(f'ERROR : input must be a SMILES, SDF, MOL or gzipped SDF'
          f' or MOL, you gave {infile}  Aborting.')
    return None

  if sfx == '.smi':
    return Chem.SmilesMolSupplier(infile, titleLine=False, sanitize=False)
  else:
    try:
      if gzipped:
        inf = gzip.open(infile)
        return Chem.ForwardSDMolSupplier(inf, sanitize=False)
      else:
        return Chem.ForwardSDMolSupplier(infile, sanitize=False)
    except (OSError, FileNotFoundError):
      print(f'ERROR : failed to open file {infile}.  Not Found.')
      return None


def extractFragments(molfile):
  """
    Read the input file and return a list of frequency counts of
    radius 2 Morgan fingerprints.  Sorted in descending order of
    frequency.  First element in each tuple is the fragment number,
    second is the count.
    """
  suppl = create_mol_supplier(molfile)
  if suppl is None:
    return None

  freqDict = defaultdict(int)
  mfpgen = rdFingerprintGenerator.GetMorganGenerator(radius=2)

  for i, mol in enumerate(suppl):
    if i and not i % 1000:
      print(f'Done {i} molecules.', flush=True)
    if mol is None or not mol:
      continue
    # print(f'Next mol : {Chem.MolToSmiles(mol)}  {mol.GetProp("_Name")}', flush=True)

    sfp = mfpgen.GetSparseCountFingerprint(mol)
    for bitId in sfp.GetNonzeroElements():
      freqDict[bitId] += sfp[bitId]

  freqs = list(freqDict.items())
  freqs.sort(key=lambda f: f[1], reverse=True)

  return freqs


def plot_freqs(freqs, output_file):
  """
    Do a log frequency plot of the data.  Assume it is already sorted.
    """
  y_freqs = [f[1] for f in freqs]
  (so.Plot(x=range(len(y_freqs)),
           y=y_freqs).add(so.Line()).scale(y='log').label(x='fragment number',
                                                          y='log frequency').save(output_file))


def score_fragments(freqs):
  """
    Calculate the score for each fragment.  This is the "logarithm of
    the ratio between the actual fragment count and the number of
    fragments forming 80% of all fragments in the database"
    Which is a bit ambiguous to my reading.  Reverse-engineering the
    RDKit pickle file of scores shows that it's a log10, not natural
    log, and the denominator is 1338.
    The denominator thus seems to be summing the total frequencies
    and then finding the number of fragments needed to give a sum of
    frequencies that's 80% of that, summing in descending order of
    frequency.  For the 22M compounds of SureChEMBL as of April 2023,
    that gives a denominator of 1336.
    Returns a dict of scores, where the key is a score, and the items
    are lists of fragments with that score.  This is the original
    format used for the pickle file in the original RDKit contrib
    version of this score.  The score is taken to 4 decimal places
    and stringified.
    freqs contains tuples of fragment id and frequency, and is assumed
    already to be sorted in descending frequency.
    """
  tot_freqs = 0
  for f in freqs:
    tot_freqs += f[1]

  tot_freqs_80 = (tot_freqs * 8) // 10
  tf = 0
  for i, freq in enumerate(freqs):
    tf += freq[1]
    if tf < tot_freqs_80:
      freq_80_num = i
  print(f'tot_freqs : {tot_freqs} : tot_freqs_80 : {tot_freqs_80}'
        f'  freq_80_num = {freq_80_num}')

  scores_dict = defaultdict(list)
  for i, freq in enumerate(freqs):
    score = f'{math.log10(freq[1] / freq_80_num):.4f}'.strip()
    scores_dict[score].append(freq[0])

  return scores_dict


def save_scores_dict(scores_dict, score_file):
  """
    The pickle appears to be list of lists, each one headed by a score
    followed by the fragments with that score.
    """
  out_lists = []
  for score, frags in scores_dict.items():
    out_lists.append([score] + frags)
  # for ol in out_lists:
  #     print(f'{len(ol)} : {ol[:10]}')
  pickle.dump(out_lists, gzip.open(score_file, 'wb'))


def main(cli_args):
  args = parse_args(cli_args)
  if args.input_file is not None:
    freqs = extractFragments(args.input_file)
  else:
    print(f'reading frequencies from {args.in_freq_pickle}')
    with open(args.in_freq_pickle, 'rb') as f:
      freqs = pickle.load(f)

  if freqs is None:
    return False

  print(freqs[:10])
  print(freqs[-10:])
  num_1000 = 0
  num_1 = 0
  for f in freqs:
    if f[1] > 1000:
      num_1000 += 1
    if f[1] == 1:
      num_1 += 1
  print(f'num above 1000 : {num_1000}  num 1 : {num_1}')

  print(f'number of fragments : {len(freqs)}')
  if HAVE_SEABORN:
    plot_freqs(freqs, args.plot_file)
  else:
    print(f'No plot file {args.plot_file} because no seaborn found.')
  if args.out_freq_pickle is not None:
    with open(args.out_freq_pickle, 'wb') as f:
      pickle.dump(freqs, f)

  scores_dict = score_fragments(freqs)
  save_scores_dict(scores_dict, args.score_file)

  return True


if __name__ == '__main__':
  sys.exit(not main(sys.argv[1:]))


#
#  Copyright (c) 2023, Glysade LLC
#  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#     * Neither the name of Novartis Institutes for BioMedical Research Inc.
#       nor the names of its contributors may be used to endorse or promote
#       products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#



"""
Glare Algorithm

Some nomenclature:

 GLARE: A New Approach for Filtering Large Reagent Lists in 
           Combinatorial Library Design Using Product Properties
    Jean-Francois Truchon* and Christopher I. Bayly

    http://pubs.acs.org/doi/pdf/10.1021/ci0504871

 A Libary is made of RGroups
 RGroups are a collection of sidechains (the paper uses Fragments)
  that can populate the rgroup position.

 We desire to optimize the Library so that we have a good chance
  of making the products in the desired property range.

 Example From the testing code, using Fake data:

    r1 = RGroups(makeFakeSidechains("aldehydes", num=1000))
    r2 = RGroups(makeFakeSidechains("boronic_acids", num=1500))
    
    lib = Library([r1,r2])
    props = [
        Property("mw",    propIdx=0, minValue=0, maxValue=500),
        Property("alogp", propIdx=1, minValue=-2.4, maxValue=5),
        Property("tpsa",  propIdx=2, minValue=0, maxValue=90)
    ]
    
    glare = Glare()
    # optimize the library...
    glare.optimize(lib, props)
    for reactant_idx, rgroup in enumerate(lib.rgroups):
        print(f"Reactants for reactant {reactant_idx}")
        for reactant in rgroup.sidechhains:
            print(reactant.name)
"""
import itertools
import math
import operator
import random
from functools import reduce


class Property:

  def __init__(self, name, propIdx, minValue, maxValue, scaffoldoffset=0.0):
    """name, propIdx, minValue, maxValue, scaffoldoffset -> initial a Property
        name is the name of the property.
         propIdx:  the index of the property in the property vector
         minValue: the minimum acceptable value for the property
         maxValue: the maximum acceptable value for the property
         scaffoldoffset: any offset from the reaction scaffold (defaults to 0)
        """
    self.name = name
    self.propIdx = propIdx
    self.minValue = minValue
    self.maxValue = maxValue
    self.offset = scaffoldoffset

  def evaluate(self, sidechains):
    """sidechains -> Evaluate a list of sidechains to see if they
        pass the property values.

        Each sidechain must have a property vector e.g. (s.props for s in sidechains)
        which is a vector of values where s.props[propIdx] is the property
        being inspected
        """
    product = self.offset
    propIdx = self.propIdx
    for s in sidechains:
      product += s.props[propIdx]
    return self.minValue <= product <= self.maxValue


class Sidechain:
  """Holds the name (identifier) and property list for the
    given sidechain/fragment.  Properties are assumed to
    be numerical values"""

  def __init__(self, name, props, goodCount=0, **extra_data):
    """name, props, goodCount=0 -> initialize a Sidechain
        initialize a sidechain.
        name: the unique name for the sidechain
        props: the property vector (see Properties class for details)
        goodCount: the number of times this reagent belongs to
            a good product, where good is a product that is in the desired
            property space.
        """
    self.name = name
    self.props = props
    self.good_count = goodCount  # shared variable
    self.dropped = False  # shared variable
    self.extra_data = extra_data

  def data(self):
    return self.extra_data

  def __str__(self):
    return "Sidechain %s(%s, goodCount=%s, **%r)" % (self.name, self.props, self.good_count,
                                                     self.extra_data)

  def __repr__(self):
    return "Sidechain(%r, %r, %s, **%r)" % (self.name, self.props, self.good_count, self.extra_data)


class RGroups:
  """Holds a collection of sidechains for the given RGroup"""

  def __init__(self, sidechains):
    """Sidechains -> RGroups
         sidechains: the list of Sidechains that make up the potential
                     sidechains at this rgroup position"""
    self.sidechains = sidechains

    self.rejected = []  # list of rejected sidechains
    self.initial_size = len(sidechains)

  def count(self):
    """Returns the number of possible sidechains"""
    return len(self.sidechains)

  def randomize(self):
    """Randomly shuffles the sidechains and reset the goodness counts"""
    random.shuffle(self.sidechains)
    for s in self.sidechains:
      s.good_count = 0

  def effectiveness(self):
    """-> return the current effectiveness of this collection
        effectiveness is the number of items left divided by the 
        initial amount"""

    return len(self.sidechains) / float(self.initial_size)

  def chunk_size(self, num_chunks):
    """num_chunks -> return the number of sidechains in each chunk
        if the sidechains are split into num_chunks chunks"""
    return int(math.ceil(float(len(self.sidechains)) / num_chunks))

  def chunk(self, chunk_idx, num_chunks):
    """chunk_idx, num)chunks -> RGroups
        return the chunk_idxth chunk given num_chunks total chunks"""
    assert chunk_idx >= 0 and chunk_idx < num_chunks, "%s %s" % (chunk_idx, num_chunks)

    n = self.chunk_size(num_chunks)
    return RGroups(self.sidechains[chunk_idx * n:(chunk_idx + 1) * n])

  def prune(self, fractionToKeep):
    """fractionToKeep -> Sort the sidechains from the most often 
        found if good products to the least, and keep the best 
        fractionToKeep percentage"""
    assert 0 < fractionToKeep <= 1.0, "fractionToKeep: %s" % fractionToKeep

    self.sidechains.sort(key=lambda x: x.good_count, reverse=True)
    fragment_index = int(len(self.sidechains) * fractionToKeep + 0.5)

    # update rejected set
    self.rejected += self.sidechains[fragment_index:]
    self.sidechains = self.sidechains[:fragment_index]


class Library:
  """A library is a collection of RGroups that need to be combinitorially
    combined"""

  def __init__(self, rgroups):
    """rgroups -> Initialize the Library.
        rgroups: the list of possible RGroups that is combinitorially
                 combined to make the library"""
    self.rgroups = rgroups

  def isValid(self):
    """If we have an empty set for any rgroup, return False"""
    for rg in self.rgroups:
      if len(rg.sidechains) == 0:
        return False
    return True

  def randomize(self):
    """randomize the order of the sidechains"""
    for rg in self.rgroups:
      rg.randomize()

  def getSidechainsPerPartition(self, total_num_partitions_per_rgroup):
    """total_num_partitions -> [num_fragments/partition for rgroup1, 
                                    num_fragments/partition for rgroup2]
        return the number of sidechains in a partition
        for each rgroup"""

    sizes = [(libIdx, max(rg.count() // total_num_partitions_per_rgroup, 1))
             for libIdx, rg in enumerate(self.rgroups)]

    # "optimally" apportion the partitions according the
    #  the glare paper see Appendix eq (8) and (9)
    # sort by size
    sizes.sort(key=lambda sz: sz[1])
    last_size = 1
    opt_sizes = []
    for libIdx, current_size in sizes[:-1]:
      opt_sizes.append((libIdx, current_size - (current_size % last_size)))
      last_size = current_size

    # From the Glare paper:
    # the last library size is set equal to the second to last
    # From Table 3, it is easy to understand that, if the fourth dimension
    # was split in 24 instead of 12, a factor of 2 would be gained from the
    # reduced size of the sublibraries. However, twice as many sublibraries
    # would be needed, and the net speedup would be null, hence, the decision to
    # set p4=p3. (p4 here is the last library)
    libIdx, current_size = sizes[-1]
    opt_sizes.append((libIdx, last_size))
    # back to the original library order
    opt_sizes.sort()
    res = [size for libIdx, size in opt_sizes]
    return res

  def chunk(self, num_partitions):
    """num_partitions -> [Library(..), Library(...)]

        Return new libraries that are chunks of this one.
        These are the libraries that get sampled to see of
        sidechains participate in good products.
        """
    partitions = self.getSidechainsPerPartition(num_partitions)
    max_subsets = max(partitions)
    enumeration_indices = []
    for i in range(max_subsets):
      combinations = []
      for size in partitions:
        combinations.append(i % size)
      enumeration_indices.append(combinations)

    library_sets = []
    for subset_index, combinations in enumerate(enumeration_indices):
      libs = []
      partitioned_rgroups = []
      for lib_index, libpart_index in enumerate(combinations):
        lib = self.rgroups[lib_index]
        num_chunks = partitions[lib_index]
        partitioned_rgroups.append(lib.chunk(chunk_idx=libpart_index, num_chunks=num_chunks))
      lib = Library(partitioned_rgroups)
      if lib.isValid():
        library_sets.append(lib)

    return library_sets

  def effectiveness(self):
    """-> returns the average effectiveness of this library set"""
    sum = 0.0
    for rg in self.rgroups:
      sum += rg.effectiveness()
    return sum / len(self.rgroups)

  def evaluate(self, props):
    """props -> num_good_enumerations, total_enumerations

        props: a list of Property evaluators for the fragments.

        returns the number of good enumerations and the total number of
        enumerations for this Library"""
    frags = [rg.sidechains for rg in self.rgroups]
    good = 0
    bad = 0
    for i, frag in enumerate(itertools.product(*frags)):
      for p in props:
        if not p.evaluate(frag):
          bad += 1
          break
      else:
        good += 1
        for sidechain in frag:
          sidechain.good_count += 1
    return good, i + 1


class Glare:
  """Glare Algorithm.  Implementation of

    GLARE: A New Approach for Filtering Large Reagent Lists in 
           Combinatorial Library Design Using Product Properties
    Jean-Francois Truchon* and Christopher I. Bayly

    http://pubs.acs.org/doi/pdf/10.1021/ci0504871

    Usage:
       # somehow make sidechains1/2 with props [mw, alogp, tpsa]
       r1 = RGroups(sidechains1)
       r2 = RGroups(sidechains2)
       lib = Library([r1, r2])
       props = [
         Property("mw", 0, 0, 500),
         Property("alogp", 1, -2.4, 5),
         Property("tpsa", 2, 0, 90)
       ] 

      glare = Glare()
      glare.optimize(lib, props)
    """

  def __init__(
      self,
      desiredFinalGoodness=0.95,
      maxIterations=100,
      rgroupScale=6.0,  # None if no scaling
      initialFraction=None,  #None=auto -100.,
      numPartitions=16):
    self.fractionGood = self.desiredFinalGoodness = desiredFinalGoodness
    self.maxIterations = maxIterations
    self.rgroupScale = rgroupScale

    if initialFraction is not None:
      self.initialFraction = initialFraction / 100.
    else:
      self.initialFraction = initialFraction
    self.numPartitions = numPartitions

  def optimize(self, library, props):
    """library, props
        Given a Library and the list of Propery evaluators,
        optimize the library.
        The library is modified in place by removing building blocks
        (sidechains) that are not likely to pass the property
        criteria.
        """
    # attempt to generate report like glare application
    print("------- PARAMETERS: --------------")
    print("GOOODNESS THRESHOLD : %s%%" % (self.desiredFinalGoodness * 100))
    print("MIN PARTITION SIZE : %s" % self.numPartitions)
    if self.initialFraction is None or self.initialFraction > 0.999:
      print("INITIAL FRACTION TO KEEP : AUTOMATIC")
    else:
      print("INITIAL FRACTION TO KEEP : %s%%" % (self.initialFraction * 100))
    print("Actual SIZE : %s = %s" %
          (" x ".join([str(len(rg.sidechains)) for rg in library.rgroups
                       ]), reduce(operator.mul, [len(rg.sidechains) for rg in library.rgroups])))

    running_total = 0.0
    Gt = self.desiredFinalGoodness

    for iteration in range(1, self.maxIterations + 1):
      # chunk of the total library into smaller more manageable sets
      #  and run combinatorial analysis on the sub libraries
      #  each of these records the number of times a sidechain is used
      #  in a successful enumeration which is then used to prune the
      #  library at the end
      #
      for rg in library.rgroups:
        rg.randomize()

      good = total = 0.0
      chunked_libs = library.chunk(self.numPartitions)
      # for each chunk, do the combinatorial check to see
      #  if reagents make good products
      for libidx, chunk in enumerate(chunked_libs):
        g, t = chunk.evaluate(props)
        good += g
        total += t
      running_total += total
      Gi = good / total  # current goodness

      if Gi < 1e-12:
        # I think we're done here :)
        fraction = 0.0
      elif iteration == 1:
        G0 = Gi  # Goodness at first iteration

        # the first time, use the initalFraction or a "good enough"
        #  value
        if self.initialFraction is not None:
          fraction = K0 = self.initialFraction
        else:
          # auto choose the fraction based on the current good percentage
          #  and the desired
          fraction = K0 = min(-1.1 * (Gt - G0) + 1.2, 0.9)
      else:
        # the second time, gradually eliminate reagents slowing
        #  down as the number of iterations increases
        #  see equation (5) in reference
        if abs(Gt - G0) < 1e-4:
          Ki = 1.0
        else:
          Ki = (1.0 - K0) * (Gi - G0) / (Gt - G0) + K0
        fraction = min(1.0, Ki)

      # prune the library to keep the highest occurring sidechains
      #  note that even if all sidechains are acceptable,
      #  some will always get pruned

      max_size = float(max([len(rg.sidechains) for rg in library.rgroups]))
      for rg in library.rgroups:
        scale = 1.0
        if self.rgroupScale is not None:
          # scale differently size rgroups via equation (6)  in paper
          numSidechains = len(rg.sidechains)
          numer = 1.0
          denom = 1.0 + math.exp(-self.rgroupScale * ((numSidechains / max_size) - 0.5))
          scale = numer / denom
        fraction_to_reject = (1.0 - fraction) * scale
        # keep the best fraction...
        rg.prune(1.0 - fraction_to_reject)

      print("-------------- ITERATION : %s ----------------------" % iteration)
      print("GOODNESS      : %s%%" % (Gi * 100))
      print("NUMBER EVAL   : %s" % (total))
      print("CUMUL EVAL    : %s" % (running_total))
      print("KEPT IN STEP  : %s%%" % (fraction * 100.))
      if not iteration:
        print("GOODNESS THRESHOLD : %s" % self.desiredFinalGoodness)
        print("MIN PARTITION SIZE : %s" % self.numPartitions)
        print("INITIAL FRACTION TO KEEP : ")
        if self.fractionToKeep > 0.999:
          print("AUTOMATIC")
        else:
          print("%s%%" % self.fractionToKeep)

      print("Actual SIZE : %s = %s" %
            (" x ".join([str(len(rg.sidechains)) for rg in library.rgroups
                         ]), reduce(operator.mul, [len(rg.sidechains) for rg in library.rgroups])))
      print("EFFECTIVENESS : %s%%" % (library.effectiveness() * 100.))

      # stopping criteria
      if iteration and Gi < 1e-12:
        return
      elif abs(Gi - self.desiredFinalGoodness) < 0.001 or \
           Gi > self.desiredFinalGoodness:
        return


######################################################################
# testing codes
def makeFakeProps():
  mw = random.randint(10, 500)
  alogp = random.randint(-10, 10)
  tpsa = random.randint(0, 180)
  return [mw, alogp, tpsa]


def makeFakeSidechains(lib, num):
  res = []
  for i in range(num):
    res.append(Sidechain(lib + "_" + str(i), makeFakeProps()))
  return res


def testGlare():
  a = RGroups(makeFakeSidechains("aldehydes", 1000))
  b = RGroups(makeFakeSidechains("boronic_acids", 1500))

  lib = Library([a, b])
  props = [
    Property("mw", 0, 0, 500, 230.1419),
    Property("alogp", 1, -2.4, 5, 2.212749),
    Property("tpsa", 2, 0, 90, 24.5)
  ]

  glare = Glare()
  glare.optimize(lib, props)
  # print out the selected reactants
  for reactant_idx, rgroup in enumerate(lib.rgroups):
    print(f"Reactants for reactant {reactant_idx}")
    for reactant in rgroup.sidechains:
      print(reactant.name)


if __name__ == "__main__":
  testGlare()



#
# Copyright (C) 2003-2006 greg Landrum and Rational Discovery LLC
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
""" Supplies a class for working with fingerprints from databases
#DOC

"""
import pickle

from rdkit import DataStructs
from rdkit.VLib.Node import VLibNode


class DbFpSupplier(VLibNode):
  """
      new fps come back with all additional fields from the
      database set in a "_fieldsFromDb" data member

    """

  def __init__(self, dbResults, fpColName='AutoFragmentFp', usePickles=True):
    """

          DbResults should be a subclass of Dbase.DbResultSet.DbResultBase

        """
    VLibNode.__init__(self)
    self._usePickles = usePickles
    self._data = dbResults
    self._fpColName = fpColName.upper()
    self._colNames = [x.upper() for x in self._data.GetColumnNames()]
    if self._fpColName not in self._colNames:
      raise ValueError('fp column name "%s" not found in result set: %s' %
                       (self._fpColName, str(self._colNames)))
    self.fpCol = self._colNames.index(self._fpColName)
    del self._colNames[self.fpCol]
    self._colNames = tuple(self._colNames)
    self._numProcessed = 0

  def GetColumnNames(self):
    return self._colNames

  def _BuildFp(self, data):
    data = list(data)
    pkl = bytes(data[self.fpCol], encoding='Latin1')
    del data[self.fpCol]
    self._numProcessed += 1
    try:
      if self._usePickles:
        newFp = pickle.loads(pkl, encoding='bytes')
      else:
        newFp = DataStructs.ExplicitBitVect(pkl)
    except Exception:
      import traceback
      traceback.print_exc()
      newFp = None
    if newFp:
      newFp._fieldsFromDb = data
    return newFp

  def next(self):
    itm = self.NextItem()
    if itm is None:
      raise StopIteration
    return itm

  __next__ = next  # py3


class ForwardDbFpSupplier(DbFpSupplier):
  """ DbFp supplier supporting only forward iteration

    >>> from rdkit import RDConfig
    >>> from rdkit.Dbase.DbConnection import DbConnect
    >>> fName = RDConfig.RDTestDatabase
    >>> conn = DbConnect(fName,'simple_combined')
    >>> suppl = ForwardDbFpSupplier(conn.GetData())

    we can loop over the supplied fingerprints:
    
    >>> fps = []
    >>> for fp in suppl:
    ...   fps.append(fp)
    >>> len(fps)
    12

    """

  def __init__(self, *args, **kwargs):
    DbFpSupplier.__init__(self, *args, **kwargs)
    self.reset()

  def reset(self):
    DbFpSupplier.reset(self)
    self._dataIter = iter(self._data)

  def NextItem(self):
    """

          NOTE: this has side effects

        """
    d = next(self._dataIter, None)
    if d is None:
      return d
    return self._BuildFp(d)


class RandomAccessDbFpSupplier(DbFpSupplier):
  """ DbFp supplier supporting random access:

  >>> import os.path
  >>> from rdkit import RDConfig
  >>> from rdkit.Dbase.DbConnection import DbConnect
  >>> fName = RDConfig.RDTestDatabase
  >>> conn = DbConnect(fName,'simple_combined')
  >>> suppl = RandomAccessDbFpSupplier(conn.GetData())
  >>> len(suppl)
  12

  we can pull individual fingerprints:

  >>> fp = suppl[5]
  >>> fp.GetNumBits()
  128
  >>> fp.GetNumOnBits()
  54

  a standard loop over the fingerprints:

  >>> fps = []
  >>> for fp in suppl:
  ...   fps.append(fp)
  >>> len(fps)
  12

  or we can use an indexed loop:

  >>> fps = [None] * len(suppl)
  >>> for i in range(len(suppl)):
  ...   fps[i] = suppl[i]
  >>> len(fps)
  12

  """

  def __init__(self, *args, **kwargs):
    DbFpSupplier.__init__(self, *args, **kwargs)
    self.reset()

  def __len__(self):
    return len(self._data)

  def __getitem__(self, idx):
    newD = self._data[idx]
    return self._BuildFp(newD)

  def reset(self):
    self._pos = -1

  def NextItem(self):
    self._pos += 1
    res = None
    if self._pos < len(self):
      res = self[self._pos]
    return res


# ------------------------------------
#
#  doctest boilerplate
#
def _runDoctests(verbose=None):  # pragma: nocover
  import doctest
  import sys
  failed, _ = doctest.testmod(optionflags=doctest.ELLIPSIS, verbose=verbose)
  sys.exit(failed)


if __name__ == '__main__':  # pragma: nocover
  _runDoctests()



#
# Copyright (C) 2003-2008 greg Landrum and Rational Discovery LLC
#
#   @@ All Rights Reserved @@
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.
#
""" contains factory class for producing signatures


"""
import copy

import numpy

from rdkit.Chem.Pharm2D import Utils
from rdkit.DataStructs import (IntSparseIntVect, LongSparseIntVect,
                               SparseBitVect)

_verbose = False


class SigFactory(object):
  """

      SigFactory's are used by creating one, setting the relevant
      parameters, then calling the GetSignature() method each time a
      signature is required.

    """

  def __init__(self, featFactory, useCounts=False, minPointCount=2, maxPointCount=3,
               shortestPathsOnly=True, includeBondOrder=False, skipFeats=None,
               trianglePruneBins=True):
    self.featFactory = featFactory
    self.useCounts = useCounts
    self.minPointCount = minPointCount
    self.maxPointCount = maxPointCount
    self.shortestPathsOnly = shortestPathsOnly
    self.includeBondOrder = includeBondOrder
    self.trianglePruneBins = trianglePruneBins
    if skipFeats is None:
      self.skipFeats = []
    else:
      self.skipFeats = skipFeats
    self._bins = None
    self.sigKlass = None

  def SetBins(self, bins):
    """ bins should be a list of 2-tuples """
    self._bins = copy.copy(bins)
    self.Init()

  def GetBins(self):
    return self._bins

  def GetNumBins(self):
    return len(self._bins)

  def GetSignature(self):
    return self.sigKlass(self._sigSize)

  def _GetBitSummaryData(self, bitIdx):
    nPts, combo, scaffold = self.GetBitInfo(bitIdx)
    fams = self.GetFeatFamilies()
    labels = [fams[x] for x in combo]
    dMat = numpy.zeros((nPts, nPts), dtype=numpy.int64)
    dVect = Utils.nPointDistDict[nPts]
    for idx in range(len(dVect)):
      i, j = dVect[idx]
      dMat[i, j] = scaffold[idx]
      dMat[j, i] = scaffold[idx]

    return nPts, combo, scaffold, labels, dMat

  def GetBitDescriptionAsText(self, bitIdx, includeBins=0, fullPage=1):
    """  returns text with a description of the bit

        **Arguments**

          - bitIdx: an integer bit index

          - includeBins: (optional) if nonzero, information about the bins will be
            included as well

          - fullPage: (optional) if nonzero, html headers and footers will
            be included (so as to make the output a complete page)

        **Returns**

          a string with the HTML

        """
    raise NotImplementedError('Missing implementation')

  def GetBitDescription(self, bitIdx):
    """  returns a text description of the bit

        **Arguments**

          - bitIdx: an integer bit index

        **Returns**

          a string

        """
    _, _, _, labels, dMat = self._GetBitSummaryData(bitIdx)
    res = " ".join(labels) + " "
    for row in dMat:
      res += "|" + " ".join([str(x) for x in row])
    res += "|"
    return res

  def _findBinIdx(self, dists, bins, scaffolds):
    """ OBSOLETE: this has been rewritten in C++
        Internal use only
         Returns the index of a bin defined by a set of distances.

        **Arguments**

          - dists: a sequence of distances (not binned)

          - bins: a sorted sequence of distance bins (2-tuples)

          - scaffolds: a list of possible scaffolds (bin combinations)

        **Returns**

          an integer bin index

        **Note**

          the value returned here is not an index in the overall
          signature.  It is, rather, an offset of a scaffold in the
          possible combinations of distance bins for a given
          proto-pharmacophore.

        """

    whichBins = [0] * len(dists)
    # This would be a ton easier if we had contiguous bins
    # i.e. if we could maintain the bins as a list of bounds)
    # because then we could use Python's bisect module.
    # Since we can't do that, we've got to do our own binary
    # search here.
    for i, dist in enumerate(dists):
      where = -1
      # do a simple binary search:
      startP, endP = 0, len(bins)
      while startP < endP:
        midP = (startP + endP) // 2
        begBin, endBin = bins[midP]
        if dist < begBin:
          endP = midP
        elif dist >= endBin:
          startP = midP + 1
        else:
          where = midP
          break
      if where < 0:
        return None
      whichBins[i] = where
    res = scaffolds.index(tuple(whichBins))
    if _verbose:
      print('----- _fBI  -----------')
      print(' scaffolds:', scaffolds)
      print(' bins:', whichBins)
      print(' res:', res)
    return res

  def GetFeatFamilies(self):
    fams = [fam for fam in self.featFactory.GetFeatureFamilies() if fam not in self.skipFeats]
    fams.sort()
    return fams

  def GetMolFeats(self, mol):
    featFamilies = self.GetFeatFamilies()
    featMatches = {}
    for fam in featFamilies:
      feats = self.featFactory.GetFeaturesForMol(mol, includeOnly=fam)
      featMatches[fam] = [feat.GetAtomIds() for feat in feats]

    return [featMatches[x] for x in featFamilies]

  def GetBitIdx(self, featIndices, dists, sortIndices=True):
    """ returns the index for a pharmacophore described using a set of
          feature indices and distances

        **Arguments***

          - featIndices: a sequence of feature indices

          - dists: a sequence of distance between the features, only the
            unique distances should be included, and they should be in the
            order defined in Utils.

          - sortIndices : sort the indices

        **Returns**

          the integer bit index

        """
    nPoints = len(featIndices)
    if nPoints > 3:
      raise NotImplementedError('>3 points not supported')
    if nPoints < self.minPointCount:
      raise IndexError('bad number of points')
    if nPoints > self.maxPointCount:
      raise IndexError('bad number of points')

    # this is the start of the nPoint-point pharmacophores
    startIdx = self._starts[nPoints]

    #
    # now we need to map the pattern indices to an offset from startIdx
    #
    if sortIndices:
      tmp = list(featIndices)
      tmp.sort()
      featIndices = tmp

    if featIndices[0] < 0:
      raise IndexError('bad feature index')
    if max(featIndices) >= self._nFeats:
      raise IndexError('bad feature index')

    if nPoints == 3:
      featIndices, dists = Utils.OrderTriangle(featIndices, dists)

    offset = Utils.CountUpTo(self._nFeats, nPoints, featIndices)
    if _verbose:
      print(f'offset for feature {str(featIndices)}: {offset}')
    offset *= len(self._scaffolds[len(dists)])

    try:
      if _verbose:
        print('>>>>>>>>>>>>>>>>>>>>>>>')
        print('\tScaffolds:', repr(self._scaffolds[len(dists)]), type(self._scaffolds[len(dists)]))
        print('\tDists:', repr(dists), type(dists))
        print('\tbins:', repr(self._bins), type(self._bins))
      bin_ = self._findBinIdx(dists, self._bins, self._scaffolds[len(dists)])
    except ValueError:
      fams = self.GetFeatFamilies()
      fams = [fams[x] for x in featIndices]
      raise IndexError('distance bin not found: feats: %s; dists=%s; bins=%s; scaffolds: %s' %
                       (fams, dists, self._bins, self._scaffolds))

    return startIdx + offset + bin_

  def GetBitInfo(self, idx):
    """ returns information about the given bit

         **Arguments**

           - idx: the bit index to be considered

         **Returns**

           a 3-tuple:

             1) the number of points in the pharmacophore

             2) the proto-pharmacophore (tuple of pattern indices)

             3) the scaffold (tuple of distance indices)

        """
    if idx >= self._sigSize:
      raise IndexError(f'bad index ({idx}) queried. {self._sigSize} is the max')
    # first figure out how many points are in the p'cophore
    nPts = self.minPointCount
    while nPts < self.maxPointCount and self._starts[nPts + 1] <= idx:
      nPts += 1

    # how far are we in from the start point?
    offsetFromStart = idx - self._starts[nPts]
    if _verbose:
      print(f'\t {nPts} Points, {offsetFromStart} offset')

    # lookup the number of scaffolds
    nDists = len(Utils.nPointDistDict[nPts])
    scaffolds = self._scaffolds[nDists]

    nScaffolds = len(scaffolds)

    # figure out to which proto-pharmacophore we belong:
    protoIdx = offsetFromStart // nScaffolds
    indexCombos = Utils.GetIndexCombinations(self._nFeats, nPts)
    combo = tuple(indexCombos[protoIdx])
    if _verbose:
      print(f'\t combo: {str(combo)}')

    # and which scaffold:
    scaffoldIdx = offsetFromStart % nScaffolds
    scaffold = scaffolds[scaffoldIdx]
    if _verbose:
      print(f'\t scaffold: {str(scaffold)}')
    return nPts, combo, scaffold

  def Init(self):
    """ Initializes internal parameters.  This **must** be called after
          making any changes to the signature parameters

        """
    accum = 0
    self._scaffolds = [0] * (len(Utils.nPointDistDict[self.maxPointCount + 1]))
    self._starts = {}
    if not self.skipFeats:
      self._nFeats = len(self.featFactory.GetFeatureFamilies())
    else:
      self._nFeats = 0
      for fam in self.featFactory.GetFeatureFamilies():
        if fam not in self.skipFeats:
          self._nFeats += 1
    for i in range(self.minPointCount, self.maxPointCount + 1):
      self._starts[i] = accum
      nDistsHere = len(Utils.nPointDistDict[i])
      scaffoldsHere = Utils.GetPossibleScaffolds(i, self._bins,
                                                 useTriangleInequality=self.trianglePruneBins)
      self._scaffolds[nDistsHere] = scaffoldsHere
      accum += (Utils.NumCombinations(self._nFeats, i) * len(scaffoldsHere))
    self._sigSize = accum
    if not self.useCounts:
      self.sigKlass = SparseBitVect
    elif self._sigSize < 2**31:
      self.sigKlass = IntSparseIntVect
    else:
      self.sigKlass = LongSparseIntVect

  def GetSigSize(self):
    return self._sigSize


try:
  from rdkit.Chem.Pharmacophores import cUtils
except ImportError:
  pass
else:
  SigFactory._findBinIdx = cUtils.FindBinIdx



#
#  Original authors: Richard Hall and Guillaume Godin
#  This file is part of the RDKit.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the RDKit source tree.

from collections import namedtuple

#
#
# Richard hall 2017
# IFG main code
# Guillaume Godin 2017
# refine output function
# astex_ifg: identify functional groups a la Ertl, J. Cheminform (2017) 9:36
from rdkit import Chem


def merge(mol, marked, aset):
  bset = set()
  for idx in aset:
    atom = mol.GetAtomWithIdx(idx)
    for nbr in atom.GetNeighbors():
      jdx = nbr.GetIdx()
      if jdx in marked:
        marked.remove(jdx)
        bset.add(jdx)
  if not bset:
    return
  merge(mol, marked, bset)
  aset.update(bset)


# atoms connected by non-aromatic double or triple bond to any heteroatom
# c=O should not match (see fig1, box 15).  I think using A instead of * should sort that out?
PATT_DOUBLE_TRIPLE = Chem.MolFromSmarts('A=,#[!#6]')
# atoms in non aromatic carbon-carbon double or triple bonds
PATT_CC_DOUBLE_TRIPLE = Chem.MolFromSmarts('C=,#C')
# acetal carbons, i.e. sp3 carbons connected to tow or more oxygens, nitrogens or sulfurs; these O, N or S atoms must have only single bonds
PATT_ACETAL = Chem.MolFromSmarts('[CX4](-[O,N,S])-[O,N,S]')
# all atoms in oxirane, aziridine and thiirane rings
PATT_OXIRANE_ETC = Chem.MolFromSmarts('[O,N,S]1CC1')

PATT_TUPLE = (PATT_DOUBLE_TRIPLE, PATT_CC_DOUBLE_TRIPLE, PATT_ACETAL, PATT_OXIRANE_ETC)


def identify_functional_groups(mol):
  marked = set()
  #mark all heteroatoms in a molecule, including halogens
  for atom in mol.GetAtoms():
    if atom.GetAtomicNum() not in (6, 1):  # would we ever have hydrogen?
      marked.add(atom.GetIdx())

#mark the four specific types of carbon atom
  for patt in PATT_TUPLE:
    for path in mol.GetSubstructMatches(patt):
      for atomindex in path:
        marked.add(atomindex)

#merge all connected marked atoms to a single FG
  groups = []
  while marked:
    grp = set([marked.pop()])
    merge(mol, marked, grp)
    groups.append(grp)


#extract also connected unmarked carbon atoms
  ifg = namedtuple('IFG', ['atomIds', 'atoms', 'type'])
  ifgs = []
  for g in groups:
    uca = set()
    for atomidx in g:
      for n in mol.GetAtomWithIdx(atomidx).GetNeighbors():
        if n.GetAtomicNum() == 6:
          uca.add(n.GetIdx())
    ifgs.append(
      ifg(atomIds=tuple(list(g)), atoms=Chem.MolFragmentToSmiles(mol, g, canonical=True),
          type=Chem.MolFragmentToSmiles(mol, g.union(uca), canonical=True)))
  return ifgs


def main():
  for ix, smiles in enumerate([
      'Cc1nc(NS(=O)(=O)c2ccc(N)cc2)nc(C)c1',  # fig1, 1
      'NC(=N)c1ccc(C=Cc2ccc(cc2O)C(=N)N)cc1',  # 2
      'CC(=O)Nc1nnc(s1)S(=O)(=O)N',  # 3
      'NS(=O)(=O)c1cc2c(NCNS2(=O)=O)cc1Cl',  # 4
      'CNC1=Nc2ccc(Cl)cc2C(=N(=O)C1)c3ccccc3',  # 5
      'Cc1onc(c1C(=O)NC2C3SC(C)(C)C(N3C2=O)C(=O)O)c4ccccc4',  # 6
      'Clc1ccccc1C2=NCC(=O)Nc3ccc(cc23)N(=O)=O',  # 7
      'COc1cc(cc(C(=O)NCC2CCCN2CC=C)c1OC)S(=O)(=O)N',  # 8
      'Cc1ccc(Cl)c(Nc2ccccc2C(=O)O)c1Cl',  # 9
      'Clc1ccc2Oc3ccccc3N=C(N4CCNCC4)c2c1',  # 10 - there is a discrepancy with the paper here!  I wonder if Peter has the ring as aromatic?
      'FC(F)(F)CN1C(=O)CN=C(c2ccccc2)c3cc(Cl)ccc13',  # 11
      'OCC1OC(CC1O)n2cnc3C(O)CNC=Nc32',  # 12
      'CCNC1CC(C)S(=O)(=O)c2sc(cc12)S(=O)(=O)N',  # 13
      'CC(O)C1C2C(C)C(=C(N2C1=O)C(=O)O)SC3CNC(C3)C(=O)N(C)C',  # 14
      'CC1CN(CC(C)N1)c2c(F)c(N)c3c(=O)c(cn(C4CC4)c3c2F)C(=O)O',  # 15
      'CC(=CCC1C(=O)N(N(C1=O)c2ccccc2)c3ccccc3)C',  # 16
      'Clc1ccc2N=C3NC(=O)CN3Cc2c1Cl',  # 17
      'CC(=O)NC1C(NC(=N)N)C=C(OC1C(O)C(O)CO)C(=O)O',  # 18
      'CC(O)C(O)C1CNc2nc(N)nc(O)c2N1',  # 19
      'NC1CCCCN(C1)c2c(Cl)cc3c(=O)c(cn(C4CC4)c3c2Cl)C(=O)O',  # 20
  ]):
    m = Chem.MolFromSmiles(smiles)
    fgs = identify_functional_groups(m)
    print('%2d: %d fgs' % (ix + 1, len(fgs)), fgs)


if __name__ == "__main__":
  main()



"""
molvs.cli
~~~~~~~~~

This module contains a command line interface for standardization.

:copyright: Copyright 2016 by Matt Swain.
:license: MIT, see LICENSE file for more details.

*Adapted for purposes of integration of MolVS into RDKit
"""

import argparse
import logging
import sys

from rdkit import Chem
from rdkit.Chem.MolStandardize import Standardizer, Validator

log = logging.getLogger(__name__)

FILETYPES = ['smi', 'mol', 'sdf']


class MolvsParser(argparse.ArgumentParser):

  def error(self, message):
    sys.stderr.write('Error: %s\n\n'.encode() % message)
    self.print_help()
    sys.exit(2)


def _read_mol(args):
  if args.smiles:
    return Chem.MolFromSmiles(args.smiles)
  elif args.intype in {'smi', 'smiles'
                       } or args.infile.name.endswith('smi') or args.infile.name.endswith('smiles'):
    return Chem.MolFromSmiles(args.infile.read())
  elif args.intype in {'mol', 'sdf'
                       } or args.infile.name.endswith('mol') or args.infile.name.endswith('sdf'):
    return Chem.MolFromMolBlock(args.infile.read())
  else:
    return Chem.MolFromSmiles(args.infile.read())


def _write_mol(mol, args):
  if args.outtype in {
      'smi', 'smiles'
  } or args.outfile.name.endswith('smi') or args.outfile.name.endswith('smiles'):
    args.outfile.write(Chem.MolToSmiles(mol))
    args.outfile.write('\n')
  elif args.outtype in {'mol', 'sdf'
                        } or args.outfile.name.endswith('mol') or args.outfile.name.endswith('sdf'):
    args.outfile.write(Chem.MolToMolBlock(mol))
  else:
    args.outfile.write(Chem.MolToSmiles(mol))
    args.outfile.write('\n')


def standardize_main(args):
  mol = _read_mol(args)
  s = Standardizer()
  mol = s.standardize(mol)
  _write_mol(mol, args)


def validate_main(args):
  mol = _read_mol(args)
  v = Validator()
  logs = v.validate(mol)
  for log in logs:
    args.outfile.write(log)
    args.outfile.write('\n')


if __name__ == '__main__':
  """Main function for molvs command line interface."""

  # Root options
  #    parser = MolvsParser(epilog='use "molvs <command> -h" to show help for a specific command')
  parser = MolvsParser(usage="usage: python cli.py [-h] {standardize,validate} ...")
  subparsers = parser.add_subparsers(title='Available commands')

  # Options common to all commands

  common_parser = MolvsParser(add_help=False)
  common_parser.add_argument('infile', nargs='?', help='input filename',
                             type=argparse.FileType('r'), default=sys.stdin)
  common_parser.add_argument('-i', '--intype', help='input filetype', choices=FILETYPES)
  common_parser.add_argument('-:', '--smiles', help='input SMILES instead of file',
                             metavar='<smiles>')
  common_parser.add_argument('-O', '--outfile', help='output filename', type=argparse.FileType('w'),
                             default=sys.stdout, metavar='<outfile>')

  # Standardize options
  standardize_parser = subparsers.add_parser('standardize', help='standardize a molecule',
                                             parents=[common_parser])
  standardize_parser.add_argument('-o', '--outtype', help='output filetype', choices=FILETYPES)
  standardize_parser.set_defaults(func=standardize_main)

  # Validate options
  validate_parser = subparsers.add_parser('validate', help='validate a molecule',
                                          parents=[common_parser])
  validate_parser.set_defaults(func=validate_main)

  args = parser.parse_args()
  try:
    args.func(args)
  except Exception as e:
    sys.stderr.write('Error: %s\n\n'.encode() % e.message)
    parser.print_help()
    sys.exit(2)



import logging
import sys

# Need to import rdBase to properly wrap exceptions
# otherwise they will leak memory
from . import rdBase

try:
  from .rdBase import rdkitVersion as __version__
except ImportError:
  __version__ = 'Unknown'
  raise

logger = logging.getLogger("rdkit")

# if we are running in a jupyter notebook, enable the extensions
try:
  kernel_name = get_ipython().__class__.__name__
  module_name = get_ipython().__class__.__module__

  if kernel_name == 'ZMQInteractiveShell' or module_name == 'google.colab._shell':
    logger.info("Enabling RDKit %s jupyter extensions" % __version__)
    from rdkit.Chem.Draw import IPythonConsole
    rdBase.LogToPythonStderr()
except Exception:
  pass

# Do logging setup at the end, so users can suppress the
# "enabling jupyter" message at the root logger.
log_handler = logging.StreamHandler(sys.stderr)
logger.addHandler(log_handler)
logger.setLevel(logging.WARN)
logger.propagate = False

# Uncomment this to use Python logging by default:
# rdBase.LogToPythonLogger()

#
#  Boost vector iterators are VERY slow, but indexing is quite fast
#   this wraps the iterators into a much faster version
#

class VectIter:
    def __init__(self, vect):
        self.vect = vect
        self.l = len(vect)
        self.pos = -1

    def __iter__(self):
        return self
    
    def __next__(self):
        self.pos += 1
        if self.pos < self.l:
            return self.vect[self.pos]
        else:
            raise StopIteration()

def __vect__iter__(vect):
    return VectIter(vect)

VECT_WRAPS = {'MatchTypeVect', 'UnsignedLong_Vect', 'VectSizeT', 'VectorOfStringVectors'}

for name, object in vars(rdBase).items():
  if name.startswith("_list") or name.startswith("_vect") or name in VECT_WRAPS:
    if hasattr(object, "__iter__"):
      object.__iter__ = __vect__iter__

